/**/

#include "i2c.h"

/**************************
Minimal i2c library
The wire library uses interrupts, which interferes with other time-critical functions 
(in this case the readout of a quadrature encoder). This minimal library seems to work  better.

The SDA and SCL channels must be defined by the user (initial values were A4, A5). 

Use as follows:
  i2c_start();
  while(!i2c_write(0x2E)){ }	//write 0x2E into register 0xF4
  i2c_stop();

  i2c_start(); // I2C restart
  while(!i2c_write(bmp_adress+0x1)){}
  long ut=ToLong(0,0,i2c_read(1),i2c_read(0));
  i2c_stop();
***************************/

void i2c_init(byte sda, byte scl) {
    _sda = sda;
    _scl = scl;
    pinMode(_sda,OUTPUT);	  	// init pins for I2C communication
    pinMode(_scl,OUTPUT);
    digitalWrite(_sda,HIGH);		// pull-up pins for two-way communication
    digitalWrite(_scl,HIGH);
}

// The following code for I2C communication is mostly from 
// http://www.robot-electronics.co.uk/acatalog/I2C_Tutorial.html
// which explains the protocol very well

void i2c_start(){				// i2c start bit sequence
    digitalWrite(_sda,HIGH);
    delayMicroseconds(I2C_WAIT);
    digitalWrite(_scl,HIGH);
    delayMicroseconds(I2C_WAIT);
    digitalWrite(_sda,LOW);
    delayMicroseconds(I2C_WAIT);
    digitalWrite(_scl,LOW);
    delayMicroseconds(I2C_WAIT);
}

void i2c_stop(){				// i2c stop bit sequence
    digitalWrite(_sda,LOW);
    delayMicroseconds(I2C_WAIT);
    digitalWrite(_scl, HIGH);
    delayMicroseconds(I2C_WAIT);
    digitalWrite(_sda,HIGH);
    delayMicroseconds(I2C_WAIT);
}

byte i2c_read(boolean ack){	// read a byte
    byte x, d=0;
    pinMode(_sda,INPUT);
    for(x=0;x<8;x++){			// read the byte d
        d<<=1;
        digitalWrite(_scl,HIGH);			// toggle clock high
        delayMicroseconds(I2C_WAIT);
        if(digitalRead(_sda)==HIGH) d+= 1;	// read bit
       digitalWrite(_scl,LOW);			// toggle clock low
    }
    pinMode(_sda,OUTPUT);

    if(ack) 
        digitalWrite(_sda,LOW);		// acknowledge the last byte read
    else digitalWrite(_sda,HIGH); 	// don't acknowledge the last byte read

    digitalWrite(_scl,HIGH);
    delayMicroseconds(I2C_WAIT);
    digitalWrite(_scl,LOW);
    digitalWrite(_sda,HIGH);
    return d;
}

byte i2c_write(byte d){	// write a byte
    byte x, b=0;
    for(x=8;x;x--){				// write the byte d
        if(d&0x80)				// write a bit
            digitalWrite(_sda,HIGH);
        else
            digitalWrite(_sda,LOW);
        digitalWrite(_scl,HIGH);	// toggle clock
        d<<=1;
        digitalWrite(_scl,LOW);
    }
    pinMode(_sda,INPUT);
    digitalWrite(_scl,HIGH);
    delayMicroseconds(I2C_WAIT);
    if(digitalRead(_sda)==LOW)
        b=1;                			// We got an ACK bit
    digitalWrite(_scl,LOW);
    pinMode(_sda,OUTPUT);
    return b;
}
